#!/usr/bin/env python
# coding=utf-8

import torch
import torch.nn as nn
import torch.nn.functional as F

from tasks.config import cfg

from models.vgg16 import VGG16Backbone
from models.resnet_ws import Resnet18WS
import utils.vgg_weights_helper as vgg_utils

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class Classifier(nn.Module):
    def __init__(self, dim_in, num_classes=20):
        super().__init__()
        self.dim_in = dim_in
        self.dim_out = hidden_dim = 4096

        self.avg_pool = nn.AdaptiveAvgPool2d((7, 7))
        self.fc1 = nn.Linear(dim_in * 7 ** 2, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, 512)
        self.fc4 = nn.Linear(512, num_classes)

        self.sigmoid = nn.Sigmoid()

    def _init_modules(self):
        nn.init.normal_(self.fc3.weight, std=0.01)
        nn.init.constant_(self.fc3.bias, 0)
        nn.init.normal_(self.fc4.weight, std=0.01)
        nn.init.constant_(self.fc4.bias, 0)

    def detectron_weight_mapping(self):
        detectron_weight_mapping = {
            'fc1.weight': 'fc6_w',
            'fc1.bias': 'fc6_b',
            'fc2.weight': 'fc7_w',
            'fc2.bias': 'fc7_b'
        }
        return detectron_weight_mapping, []

    def forward(self, x):
        x = self.avg_pool(x)  # [1,512,7,7]
        batch_size = x.size(0)
        x = F.relu(self.fc1(x.view(batch_size, -1)), inplace=True)  # [1,4096]
        x = F.relu(self.fc2(F.dropout(x, 0.2)), inplace=True)  # [1,4096]
        x = F.relu(self.fc3(F.dropout(x, 0.2)), inplace=True)  # [1, 512]
        x = F.relu(self.fc4(F.dropout(x, 0.2)), inplace=True)  # [1, num_classes]
        x = self.sigmoid(x)

        return x


class ClassificationModel(nn.Module):
    def __init__(self, num_classes=20, input_channels=3, refine_times=3, backbone='vgg16'):
        super().__init__()

        self.mapping_to_detectron = None
        self.orphans_in_detectron = None
        self.refine_times = refine_times

        # Backbone for feature extraction
        self.backbone_name = backbone
        if backbone == 'vgg16':
            self.backbone = VGG16Backbone(input_channels=input_channels, freeze_at=0)
        elif backbone == 'resnet-ws':
            self.backbone = Resnet18WS(input_channels=input_channels)

        self.classifier = Classifier(self.backbone.dim_out, num_classes)
        self.multi_label_loss = nn.MultiLabelSoftMarginLoss()

        self.inner_iter = 0
        self._init_modules()
        if input_channels != 3 and self.backbone_name == 'vgg16':
            with torch.no_grad():
                first_layer_pretrained_weights = self.backbone.conv1[0].weight.clone()
                self.backbone.conv1[0] = nn.Conv2d(input_channels, 64, kernel_size=3, stride=1, padding=1, bias=True)
                self.backbone.conv1[0].weight[:, :3, :, :] = torch.nn.Parameter(first_layer_pretrained_weights)
                self.backbone.conv1[0].weight[:, 3:, :, :] = torch.nn.Parameter(first_layer_pretrained_weights)
                self.backbone.conv1[0].requires_grad = False

    def _init_modules(self):
        if cfg.MODEL.LOAD_IMAGENET_PRETRAINED_WEIGHTS and self.backbone_name == 'vgg16':
            vgg_utils.load_pretrained_imagenet_weights(self)

        if cfg.TRAIN.FREEZE_CONV_BODY:
            for p in self.Conv_Body.parameters():
                p.requires_grad = False

    def set_inner_iter(self, inner_iter):
        self.inner_iter = inner_iter

    def forward(self, data, rois, labels):
        with torch.set_grad_enabled(self.training):

            backbone_feat = self.backbone(data)
            cls_score = self.classifier(backbone_feat)

            return_dict = {}
            if self.training:

                return_dict['losses'] = {}

                # image classification loss
                loss_im_cls = self.multi_label_loss(cls_score, labels)
                return_dict['losses']['loss_im_cls'] = loss_im_cls

                # pytorch0.4 bug on gathering scalar(0-dim) tensors
                for k, v in return_dict['losses'].items():
                    return_dict['losses'][k] = v.unsqueeze(0)

            cls_pred = (cls_score >= 0.5).int()
            return_dict['classification'] = {}
            incorrect_count = torch.abs(cls_pred - labels).sum().detach().cpu()
            return_dict['classification']['incorrect_count'] = int(incorrect_count)
            return_dict['classification']['correct_count'] = int(labels.shape[1] - incorrect_count)
            return_dict['classification']['total_count'] = int(labels.shape[1])
            return return_dict

    @property
    def detectron_weight_mapping(self):
        if self.mapping_to_detectron is None:
            d_wmap = {}  # detectron_weight_mapping
            d_orphan = []  # detectron orphan weight list
            for name, m_child in self.named_children():
                if list(m_child.parameters()):  # if module has any parameter
                    child_map, child_orphan = m_child.detectron_weight_mapping()
                    d_orphan.extend(child_orphan)
                    for key, value in child_map.items():
                        new_key = name + '.' + key
                        d_wmap[new_key] = value
            self.mapping_to_detectron = d_wmap
            self.orphans_in_detectron = d_orphan

        return self.mapping_to_detectron, self.orphans_in_detectron


def loot_model(args, num_classes=20, input_channels=3, backbone='vgg16'):
    print("Using model description:", args.model)
    model = ClassificationModel(num_classes=num_classes, input_channels=input_channels, backbone=backbone)
    return model

